import os
import time
import psutil

from airflow.stats import Stats
from airflow.operators.python import PythonOperator
from airflow.models.baseoperator import chain

from datetime import datetime, timedelta


def stats_metric(task_id: str, status: str, *args):
    """
        Высчитывает метрики потребления для тасок и публикует в statsD
    """
    p = psutil.Process(os.getpid())
    cpu_times = p.cpu_times()            # user/system CPU seconds
    mem_info  = p.memory_info()          # rss, vms и т.п.
    io_counters = p.io_counters()        # read_bytes, write_bytes и т.п.

    tag_into = {"task_id":task_id, "phase":status, "dag_id":dags_id}

    metric_cpu_times_user = f"airflow_metric_{task_id}_cpu_times_user_{status} ({cpu_times.user})"
    metric_cpu_times_system = f"airflow_metric_{task_id}_cpu_times_system_{status} ({cpu_times.system})"
    metric_mem_info_rss = f"airflow_metric_{task_id}_mem_info_rss_{status} ({mem_info.rss})"
    metric_mem_info_vms = f"airflow_metric_{task_id}_mem_info_vms_{status} ({mem_info.vms})"
    metric_io_counters_read_bytes = f"airflow_metric_{task_id}_io_counters_read_bytes_{status} ({io_counters.read_bytes})"
    metric_io_counters_write_bytes = f"airflow_metric_{task_id}_io_counters_write_bytes_{status} ({io_counters.write_bytes})"


    Stats.gauge(f'metric.{task_id}.cpu_times_user', cpu_times.user, tags=tag_into)
    Stats.gauge(f'metric.{task_id}.cpu_times_system', cpu_times.system, tags=tag_into)
    Stats.gauge(f'metric.{task_id}.mem_info_rss', mem_info.rss, tags=tag_into)
    Stats.gauge(f'metric.{task_id}.mem_info_vms', mem_info.vms, tags=tag_into)
    Stats.gauge(f'metric.{task_id}.io_counters_read_bytes', io_counters.read_bytes, tags=tag_into)
    Stats.gauge(f'metric.{task_id}.io_counters_write_bytes', io_counters.write_bytes, tags=tag_into)

    print(f"""{metric_cpu_times_user}\n
        {metric_cpu_times_system}\n
        {metric_mem_info_rss}\n
        {metric_mem_info_vms}\n
        {metric_io_counters_read_bytes}\n
        {metric_io_counters_write_bytes}""")


def chain_with_intermediates(tasks):
    """
    Вставляет между каждой парой тасок список тасок c метриками.
    Затем разворачивает все в единую цепочку зависимостей.
    [MonitoringOperator_start, BaseOperator, MonitoringOperator_finish] — примерный формат
    """

    chain_list = []
    
    for task in tasks:
 
        pre  = PythonOperator(task_id=f"airflow_monitoring_{task.task_id}_start", 
                              python_callable=stats_metric, 
                              op_args=[task.task_id, 'start'])
        
        post = PythonOperator(task_id=f"airflow_monitoring_{task.task_id}_finish", 
                              python_callable=stats_metric, 
                              op_args=[task.task_id, 'finish'])
        
        chain_list.append(pre)
        chain_list.append(task)
        chain_list.append(post)

    chain(*chain_list)
    print(*chain_list)


from abc import abstractmethod
from datetime import datetime, timedelta, timezone
from typing import Any
from zoneinfo import ZoneInfo

import pendulum
from airflow.models import BaseOperator
from airflow.sensors.base import BaseSensorOperator
from airflow.utils.context import Context


class AirflowBaseSensor(BaseSensorOperator):
    def __init__(self, **kwargs):
        super().__init__(mode="reschedule", **kwargs)
        # only for info purpose
        self._dag_id: str = None

    def __init_context(self, context: Context):
        self._dag_id = context["dag"].dag_id
        self.execution_date = context["execution_date"]
        self.execution_date_moscow_tz = self.execution_date.astimezone(
            pendulum.timezone("Europe/Moscow")
        )
        self.execution_date_moscow = (
            self.execution_date + self.execution_date_moscow_tz.utcoffset()
        )
        self.task_instance = context["task_instance"]

    def _poke(self) -> Any:
        raise NotImplementedError

    def poke(self, context):
        self.__init_context(context)
        output = self._poke()
        self.task_instance.xcom_push(key="sensor_data", value=output)
        return output


class AirflowBaseOperator(BaseOperator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sensor_metadata: Any = None
        # only for info purpose
        self._dag_id: str = None

    def __init_context(self, context: Context):
        self._dag_id = context["dag"].dag_id
        self.execution_date = context["execution_date"]
        self.execution_date_moscow_tz = self.execution_date.astimezone(
            pendulum.timezone("Europe/Moscow")
        )
        self.execution_date_moscow = (
            self.execution_date + self.execution_date_moscow_tz.utcoffset()
        )

        self.task_instance = context["task_instance"]
        self.upstream_metadata = {}
        for task_id in self.upstream_task_ids:
            self.upstream_metadata[task_id] = self.task_instance.xcom_pull(
                task_ids=task_id, key="task_data"
            )
            self.sensor_metadata = self.task_instance.xcom_pull(
                task_ids=task_id, key="sensor_data"
            )

    @abstractmethod
    def _execute(self) -> Any:
        raise NotImplementedError

    def execute(self, context: Context) -> Any:
        self.__init_context(context)
        data = self._execute()
        self.task_instance.xcom_push(key="task_data", value=data)

        return data



import logging
import os
import warnings
from typing import Dict, Optional

import boto3
import pandas as pd
from botocore.config import Config
from pandas import DataFrame as PandasDataFrame

from dump.config_utils import load_config
from dump.files.tmp_util import TemporaryFileSystem


ZIP_FORMATS_DICT = {"zip": "zip", "orc": "orc", "gz": "gzip"}


class S3Client:
    def __init__(
        self, creds_section: str, config_filiname: str = "s3_config.ini"
    ) -> None:
        self.creds_section = creds_section

        self.config = load_config(filename=config_filiname, section=creds_section)
        self.conn_config = Config(
            s3={
                "addressing_style": "virtual",
            },
            retries={"max_attempts": 10, "mode": "standard"},
            region_name="us-east-1",
        )

        self.client = boto3.client(
            "s3",
            **self.config,
            config=self.conn_config,
        )

    def get_files_info(self, bucket_name: str, object_folder: Optional[str] = ""):
        files_dict = {}
        paginator = self.client.get_paginator("list_objects_v2")
        pages = paginator.paginate(Bucket=bucket_name, Prefix=object_folder)
        for response in pages:
            if "Contents" in response:
                for obj in response["Contents"]:
                    size = obj["Size"]
                    files_dict[obj["Key"]] = {
                        "size_bytes": size,
                        "size_gb": size / 1024**3,
                    }
        return files_dict


class S3Upload(S3Client):
    """
    To upload files to s3
    """

    def __init__(
        self,
        bucket_name: str,
        object_folder: str = "",
        creds_section: str = "test",
    ) -> None:
        super().__init__(creds_section)
        self.bucket_name = bucket_name
        self.object_folder = object_folder

    def _get_path(self, file_name: str) -> str:
        """
        creating path where to save file in s3
        """
        path = os.path.join(self.object_folder, file_name)
        return path

    def upload_to_s3(self, tmp_path_file: str, file_name: str):
        """
        tmp_path_file: str - file location that we want to save
        file_name: str - new file name that will be displayed in s3

        upload local file into s3 bucket with new name 'file_name'
        """
        path = self._get_path(file_name)
        self.client.upload_file(tmp_path_file, self.bucket_name, path)


class S3UploadDF(S3Upload):
    def __init__(
        self,
        df_compression=False,
        df_compression_format: str = "gz",
        orc_compession_engine: str = "ZLIB",
        compresslevel: int = 4,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.df_compression = df_compression
        self.df_compression_format = df_compression_format
        self.orc_compession_engine = orc_compession_engine
        self.compresslevel = compresslevel

        if self.df_compression:
            if self.df_compression_format not in ZIP_FORMATS_DICT.keys():
                raise ValueError(
                    f"Unkowm df_compression_format: {self.df_compression_format}, avaliable formats: {ZIP_FORMATS_DICT.keys()}"
                )
            if self.df_compression_format == "orc":
                tmp_file_format = f"orc"
            else:
                tmp_file_format = f"csv.{self.df_compression_format}"

            self._tmp_file_system = TemporaryFileSystem(file_format=tmp_file_format)
        else:
            self._tmp_file_system = TemporaryFileSystem(file_format="csv")

    def __del__(self):
        self._tmp_file_system.cleanup()

    def __get_file_name_no_ext(self, file_name: str):
        splited_file = file_name.split(".")
        extension = splited_file[-1]

        if extension in ZIP_FORMATS_DICT:
            logging.info(
                "In filename already specified extension, that will be ignored."
            )

        return splited_file[0]

    def upload_to_s3_df(self, df: PandasDataFrame, file_name: str):
        save_path = self._tmp_file_system.save_path()
        file_extension = file_name.split(".")[-1]

        if not self.df_compression:
            # if  doesnt need compression
            if file_extension in ZIP_FORMATS_DICT:
                raise ValueError(
                    f"You pass filename contains: '{file_extension}' format in it. U should pass df_compression=True"
                )
            df.to_csv(save_path, index=False)
        elif self.df_compression_format != "orc":
            file_name_no_ext = self.__get_file_name_no_ext(file_name)
            file_name = f"{file_name_no_ext}.csv.{self.df_compression_format}"

            compression_opts = dict(
                method=ZIP_FORMATS_DICT.get(self.df_compression_format),
                compresslevel=self.compresslevel,
            )
            df.to_csv(save_path, index=False, compression=compression_opts)
        elif self.df_compression_format == "orc":
            file_name_no_ext = self.__get_file_name_no_ext(file_name)
            file_name = f"{file_name_no_ext}.{self.df_compression_format}"

            compression_opts = dict(compression=self.orc_compession_engine)
            df.to_orc(save_path, engine_kwargs=compression_opts)

        logging.info(f"uploading to s3, path: {save_path}")
        logging.info(f"uploading to s3, filename: {file_name}")
        self.upload_to_s3(save_path, file_name)


class S3Download(S3Client):
    def __init__(
        self,
        bucket_name: str,
        object_folder: str = "",
        creds_section: str = "test",
    ) -> None:
        super().__init__(creds_section)
        self.bucket_name = bucket_name
        self.object_folder = object_folder

        self._tmp_file_system = TemporaryFileSystem()

        # requires for not download same file more than one time
        self.__downloads_history: Dict[str, str] = {}

    def __del__(self):
        self._tmp_file_system.cleanup()

    def _check_path_exist(self, path: str) -> bool:
        exsist_files = self.get_files_info(self.bucket_name, self.object_folder)
        return path in exsist_files

    def _get_object_path(self, file_name) -> str:
        path = os.path.join(self.object_folder, file_name)
        return path

    def __validate_file_system_by_filename(self, download_file_name: str):
        extension = download_file_name.split(".")[-1]
        if extension:
            self._tmp_file_system.file_format = extension

    def download_from_s3(self, download_file_name: str) -> str:
        """
        downloading file from s3 and put the file into temporary directory
        returns path to file in temporary directory
        """
        self.__validate_file_system_by_filename(download_file_name)

        object_path = self._get_object_path(download_file_name)
        if self._check_path_exist(object_path):
            if object_path not in self.__downloads_history:
                save_path = self._tmp_file_system.save_path()
                self.client.download_file(self.bucket_name, object_path, save_path)
                self.__downloads_history[object_path] = save_path

            return self.__downloads_history[object_path]
        else:
            raise ValueError(f"The file {object_path} was not found")

    def download_from_s3_df(
        self, download_file_name: str, encoding="utf-8"
    ) -> PandasDataFrame:
        if download_file_name.split(".")[-1] == "orc":
            warnings.warn("Next time use download_from_s3_df_orc method!")
            return self.download_from_s3_df_orc(download_file_name)

        df_path = self.download_from_s3(download_file_name)
        return pd.read_csv(df_path, encoding=encoding, low_memory=False)

    def download_from_s3_df_orc(self, download_file_name: str) -> PandasDataFrame:
        df_path = self.download_from_s3(download_file_name)
        return pd.read_orc(df_path)



import logging

import pymysql

from dump.config_utils import load_config


TYPE_MAPPER = {
    # Integer types
    "tinyint": "Int16",
    "smallint": "Int32",
    "mediumint": "Int32",
    "int": "Int64",
    "integer": "Int64",
    "bigint": "Int64",
    # Float types
    "float": "float32",
    "real": "float64",
    "double": "float64",
    "decimal": "float64",
    "numeric": "float64",
    # Other types
    "USER-DEFINED": "str",
    "character varying": "str",
    "character": "str",
    "text": "str",
    "boolean": "bool",
    "timestamp without time zone": "timestamp",
    "timestamp": "timestamp",
    "date": "timestamp",
}


class ConnectorMySQL:
    def __init__(
        self,
        db_config_name: str = "postgres",
    ) -> None:
        self.db_config_name = db_config_name

        self.__config = load_config(section=self.db_config_name)
        self.__config["port"] = int(self.__config["port"])

        self._conn = None
        self._cursor = None

    @property
    def conn(self):
        if self._conn is None:
            try:
                self._conn = pymysql.connect(
                    **self.__config, ssl={"check_hostname": False}
                )
            except pymysql.Error as error:
                raise error
        return self._conn

    @property
    def cursor(self):
        if self._cursor is None:
            self._cursor = self.conn.cursor()
        return self._cursor

    def close(self):
        try:
            self.conn.close()
            self._conn = None
            self._cursor = None
        except Exception as e:
            logging.info(f"Error while closing connection: {e}")

    @property
    def commit(self):
        self.conn.commit()

    def execute(self, query: str):
        self.cursor.execute(query)
        self.conn.commit()

    def fetchall(self, query: str) -> list:
        self.execute(query)
        values = self.cursor.fetchall()
        self.conn.commit()

        self.cursor.close()
        self.close()
        return values


class DBUtilsMySQL(ConnectorMySQL):
    @staticmethod
    def _validate_output(inp: list):
        return [x[0] for x in inp]

    def get_table_structure(self, table_name: str) -> dict:
        query = f"""
            SELECT column_name, data_type, is_nullable, ordinal_position
            FROM information_schema.columns
            WHERE table_name = '{table_name}';
        """

        schema = {}
        for column_name, data_type, is_nullable, position in self.fetchall(query):
            orig_data_type = data_type
            if "int" in data_type and data_type not in TYPE_MAPPER.keys():
                data_type = "integer"
            if (
                "numeric" in data_type or "decimal" in data_type
            ) and data_type not in TYPE_MAPPER.keys():
                data_type = "decimal"
            if "time" in data_type:
                data_type = "timestamp"

            schema[column_name] = {
                "data_type": TYPE_MAPPER.get(data_type, "str"),
                "orig_data_type": orig_data_type,
                "is_nullable": is_nullable,
                "position": position,
            }
        return schema
